#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>

int line = 1, label = 1;

int token(int eof_acceptable) {
	int c;
	do {
		c = fgetc(stdin);
		if(c == '\n') line++;
	} while(isspace(c));
	if(c == EOF && !eof_acceptable) {
		fputs("Unexpected EOF.\n", stderr);
		exit(1);
	}
	return c;
}

int expr(void) {
	int tok;
	int donelabel, elselabel;
	int funname, paramnr;

	tok = token(0);
	if(isdigit(tok)) {
		printf("mov eax,[ebp+8+4*%d]\n", tok-'0');
		return 0;
	}
	switch(tok) {
		case '-':
			expr();
			puts("mov eax,[eax]");
			return 0;
		case '+':
			expr();
			puts("mov eax,[eax+4]");
			return 0;
		case '#':
			puts("xor eax,eax");
			return 0;
		case '[':
			expr();
			puts("push eax");
			expr();
			tok = token(0);
			if(tok != ']') {
				fprintf(stderr, "%d: \"]\" expected\n", line);
				exit(1);
			}
			puts(	"mov [edi+4],eax\n"
					"pop dword[edi]\n"
					"mov eax,edi\n"
					"lea edi,[edi+8]");
			return 0;
		case '<':
			expr();
			tok = token(0);
			if(tok != '>') {
				fprintf(stderr, "%d: \">\" expected\n", line);
				exit(1);
			}
			donelabel = label++;
			elselabel = label++;
			printf(	"test eax,eax\n"
					"jz l%d\n", elselabel);
			expr();
			printf(	"jmp l%d\n"
					"l%d:\n", donelabel, elselabel);
			expr();
			printf("l%d:\n", donelabel);
			return 0;
		default:
			paramnr = 0;
			funname = tok;
			tok = token(0);
			if(tok != '(') {
				fprintf(stderr, "%d: \"%c\" found, \"(\" expected\n",line,tok);
				exit(1);
			}
			while(tok != ')') {
				expr();
				puts("push eax");
				paramnr++;
				tok = token(0);
				if(tok != ')' && tok != ',') {
					fprintf(stderr, "%d: \",\" or \")\" expected\n", line);
					exit(1);
				}
			}
			printf(	"call fun_%02x ; %c\n"
					"add esp,4*%d\n", funname, funname, paramnr);
			return 0;
	};

}

int main(void) {
	int tok;

	printf(	"format ELF\n"
			"section '.text' executable\n\n"
			"public thai\n"
			"thai:\n"
			"push ebp\n"
			"mov ebp,esp\n"
			"push edi\n"
			"mov edi,data_area\n"
			"push dword[ebp+8]\n"
			"call fun_%02x\n"
			"add esp,4\n"
			"pop edi\n"
			"mov esp,ebp\n"
			"pop ebp\n"
			"ret\n\n", '_');

	for(;;) {
		tok = token(1);
		if(tok == ':') {
			tok = token(0);
			printf(	"fun_%02x: ; %c\n"
					"push ebp\n"
					"mov ebp,esp\n", tok, tok);
			expr();
			puts(	"mov esp,ebp\n"
					"pop ebp\n"
					"ret\n\n");
		} else if(tok == EOF) {
			puts(	"\nsection '.bss' writeable\n"
					"data_area: rd 0x100000\n");
			fprintf(stderr, "Done.\n");
			exit(0);
		} else {
			fprintf(stderr, "%d: Unexpected token: \"%c\"\n", line, tok);
			exit(1);
		}
	}
}

